from enum import IntEnum


class Squads(IntEnum):
    NO_SQUAD = 0,
    ALPHA = 1,
    BRAVO = 2,
    CHARLIE = 3,
    DELTA = 4,
    ECHO = 5,
    FOXTROT = 6,
    GOLF = 7,
    HOTEL = 8,
    INDIA = 9,
    JULIETT = 10,
    KILO = 11,
    LIMA = 12,
    MIKE = 13,
    NOVEMBER = 14,
    OSCAR = 15,
    PAPA = 16,
    QUEBEC = 17,
    ROMEO = 18,
    SIERRA = 19,
    TANGO = 20,
    UNIFORM = 21,
    WHISKEY = 22,
    XRAY = 23,
    YANKEE = 24,
    ZULU = 25,
    ASH = 26,
    BAKER = 27,
    CAST = 28,
    DIVER = 29,
    EAGLE = 30,
    FISHER = 31,
    GEORGE = 32,
    HANOVER = 33,
    ICE = 34,
    JAKE = 35,
    KING = 36,
    LASH = 37,
    MULE = 38,
    NEPTUNE = 39,
    OSTEND = 40,
    PAGE = 41,
    QUAIL = 42,
    RAFT = 43,
    SCOUT = 44,
    TARE = 45,
    UNIT = 46,
    WILLIAM = 47,
    XAINTRIE = 48,
    YOKE = 49,
    ZEBRA = 50,
    ACE = 51,
    BEER = 52,
    CAST2 = 53,
    DUFF = 54,
    EDWARD = 55,
    FREDDY = 56,
    GUSTAV = 57,
    HENRY = 58,
    IVAR = 59,
    JAZZ = 60,
    KEY = 61,
    LINCOLN = 62,
    MARY = 63,
    NORA = 64



from dataclasses import dataclass

from battlebit_community_server_api.helper.StructHelper import write_string
from battlebit_community_server_api.model.WeaponType import WeaponType


@dataclass
class Weapon:
    name: str
    weapon_type: WeaponType

    def as_bytes(self) -> bytes:
        return write_string(self.name)


ACR = Weapon("ACR", WeaponType.RIFLE)
AK15 = Weapon("AK15", WeaponType.RIFLE)
AK74 = Weapon("AK74", WeaponType.RIFLE)
G36C = Weapon("G36C", WeaponType.RIFLE)
HONEY_BADGER = Weapon("Honey Badger", WeaponType.PERSONAL_DEFENSE_WEAPON)
KRISS_VECTOR = Weapon("Kriss Vector", WeaponType.SUB_MACHINE_GUN)
L86A1 = Weapon("L86A1", WeaponType.LIGHT_SUPPORT_GUN)
L96 = Weapon("L96", WeaponType.SNIPER_RIFLE)
M4A1 = Weapon("M4A1", WeaponType.RIFLE)
M9 = Weapon("M9", WeaponType.PISTOL)
M110 = Weapon("M110", WeaponType.DMR)
M249 = Weapon("M249", WeaponType.LIGHT_MACHINE_GUN)
MK14EBR = Weapon("MK14 EBR", WeaponType.DMR)
MK20 = Weapon("MK20", WeaponType.DMR)
MP7 = Weapon("MP7", WeaponType.SUB_MACHINE_GUN)
PP2000 = Weapon("PP2000", WeaponType.SUB_MACHINE_GUN)
SCAR_H = Weapon("SCAR-H", WeaponType.RIFLE)
SSG69 = Weapon("SSG 69", WeaponType.SNIPER_RIFLE)
SV98 = Weapon("SV-98", WeaponType.SNIPER_RIFLE)
UMP45 = Weapon("UMP-45", WeaponType.SUB_MACHINE_GUN)
UNICA = Weapon("Unica", WeaponType.HEAVY_PISTOL)
USP = Weapon("USP", WeaponType.PISTOL)
AS_VAL = Weapon("As Val", WeaponType.CARBINE)
AUG_A3 = Weapon("AUG A3", WeaponType.RIFLE)
DESERT_EAGLE = Weapon("Desert Eagle", WeaponType.HEAVY_PISTOL)
FAL = Weapon("FAL", WeaponType.RIFLE)
GLOCK_18 = Weapon("Glock 18", WeaponType.AUTO_PISTOL)
M200 = Weapon("M200", WeaponType.SNIPER_RIFLE)
MP443 = Weapon("MP 443", WeaponType.PISTOL)
FAMAS = Weapon("FAMAS", WeaponType.RIFLE)
MP5 = Weapon("MP5", WeaponType.SUB_MACHINE_GUN)
P90 = Weapon("P90", WeaponType.PERSONAL_DEFENSE_WEAPON)
MSR = Weapon("MSR", WeaponType.SNIPER_RIFLE)
PP19 = Weapon("PP19", WeaponType.SUB_MACHINE_GUN)
SVD = Weapon("SVD", WeaponType.DMR)
REM_700 = Weapon("Rem700", WeaponType.SNIPER_RIFLE)
SG550 = Weapon("SG550", WeaponType.RIFLE)
GROZA = Weapon("Groza", WeaponType.PERSONAL_DEFENSE_WEAPON)
HK419 = Weapon("HK419", WeaponType.RIFLE)
SCORPION_EVO = Weapon("ScorpionEVO", WeaponType.CARBINE)
RSH_12 = Weapon("Rsh12", WeaponType.HEAVY_PISTOL)
MG36 = Weapon("MG36", WeaponType.LIGHT_SUPPORT_GUN)
AK5C = Weapon("AK5C", WeaponType.RIFLE)
ULTIMAX_100 = Weapon("Ultimax100", WeaponType.LIGHT_MACHINE_GUN)

ALL_WEAPONS = {w.name: w for w in [ACR, AK15, AK74, G36C, HONEY_BADGER, KRISS_VECTOR, L86A1, L96, M4A1, M9, M110, M249,
                                   MK14EBR, MK20, MP7, PP2000, SCAR_H, SSG69, SV98, UMP45, UNICA, USP, AS_VAL, AUG_A3,
                                   DESERT_EAGLE, FAL, GLOCK_18, M200, MP443, FAMAS, MP5, P90, MSR, PP19, SVD, REM_700,
                                   SG550, GROZA, HK419, SCORPION_EVO, RSH_12, MG36, AK5C, ULTIMAX_100]}



from dataclasses import dataclass
from typing import Optional

from battlebit_community_server_api.helper.StructHelper import *
from battlebit_community_server_api.model.Attachment import Attachment, ALL_ATTACHMENTS
from battlebit_community_server_api.model.Weapon import Weapon, ALL_WEAPONS


@dataclass
class WeaponItem:
    base_weapon: Weapon
    main_sight: Optional[Attachment]
    top_sight: Optional[Attachment]
    canted_sight: Optional[Attachment]
    barrel: Optional[Attachment]
    side_rail: Optional[Attachment]
    under_rail: Optional[Attachment]
    bolt_action: Optional[Attachment]
    skin: Optional[int]
    magazine: Optional[int]

    def as_bytes(self) -> bytes:
        out = self.base_weapon.as_bytes()
        for attachment in [self.main_sight, self.top_sight, self.canted_sight, self.barrel, self.side_rail, self.under_rail, self.bolt_action]:
            if attachment:
                out += attachment.as_bytes()
            else:
                out += struct.pack("H", 4)
                out += "none".encode("utf-8")
        out += struct.pack("B", self.skin) if self.skin else struct.pack("B", 0)
        out += struct.pack("B", self.magazine) if self.magazine else struct.pack("B", 0)
        return out


def read_from_bytes(data: bytes) -> tuple[WeaponItem, bytes]:
    """ Reads from 0 until data is completed, returns remaining bytes """
    weapon_name_size, data = read_uint16(data)
    weapon_name, data = read_string(data, weapon_name_size)
    try:
        weapon = ALL_WEAPONS[weapon_name]
    except KeyError:
        weapon = ALL_WEAPONS["M4A1"]  # ToDo: Make configurable

    main_sight_size, data = read_uint16(data)
    main_sight_name, data = read_string(data, main_sight_size)
    try:
        main_sight = ALL_ATTACHMENTS[main_sight_name]
    except KeyError:
        main_sight = None

    top_sight_size, data = read_uint16(data)
    top_sight_name, data = read_string(data, top_sight_size)
    try:
        top_sight = ALL_ATTACHMENTS[top_sight_name]
    except KeyError:
        top_sight = None

    canted_sight_size, data = read_uint16(data)
    canted_sight_name, data = read_string(data, canted_sight_size)
    try:
        canted_sight = ALL_ATTACHMENTS[canted_sight_name]
    except KeyError:
        canted_sight = None

    barrel_size, data = read_uint16(data)
    barrel_name, data = read_string(data, barrel_size)
    try:
        barrel = ALL_ATTACHMENTS[barrel_name]
    except KeyError:
        barrel = None

    side_rail_size, data = read_uint16(data)
    side_rail_name, data = read_string(data, side_rail_size)
    try:
        side_rail = ALL_ATTACHMENTS[side_rail_name]
    except KeyError:
        side_rail = None

    under_rail_size, data = read_uint16(data)
    under_rail_name, data = read_string(data, under_rail_size)
    try:
        under_rail = ALL_ATTACHMENTS[under_rail_name]
    except KeyError:
        under_rail = None

    bolt_action_size, data = read_uint16(data)
    bolt_action_name, data = read_string(data, bolt_action_size)
    try:
        bolt_action = ALL_ATTACHMENTS[bolt_action_name]
    except KeyError:
        bolt_action = None

    skin_index, data = read_uint8(data)
    magazine_index, data = read_uint8(data)

    return WeaponItem(base_weapon=weapon,
                      main_sight=main_sight,
                      top_sight=top_sight,
                      canted_sight=canted_sight,
                      barrel=barrel,
                      side_rail=side_rail,
                      under_rail=under_rail,
                      bolt_action=bolt_action,
                      skin=skin_index,
                      magazine=magazine_index
                      ), data



import logging

from battlebit_community_server_api.ApiServer import ApiServer, build_api_server
from battlebit_community_server_api.command.Command import SayToAllChat
from battlebit_community_server_api.handler.DefaultHandlers import DefaultHandlers
from battlebit_community_server_api.model.OpCodes import OpCodes
from battlebit_community_server_api.service.TcpParsingService import TcpParsingService

my_api_server: ApiServer = build_api_server("0.0.0.0", 30000)


async def my_custom_on_player_joining(data: bytes) -> bytes:
    """
    This function:

    1. Parses the data received by the "On Player Joining" event.
    2. Builds a command to greet the player in chat.
    3. Schedules the command to be executed.
    4. Executes the default handler for "On Player Joining" event.
    """
    steam_id, player_stats = TcpParsingService.parse_on_player_joining(data)
    my_new_command = SayToAllChat(f"Welcome user {steam_id}!")
    await my_api_server.add_command_to_queue(my_new_command)

    return await DefaultHandlers.get_default_handler_by_op_code(OpCodes.ON_PLAYER_JOINING)(data)


logging.basicConfig(level=logging.DEBUG)  # Optional: If you want to have an in-depth view into what's going on.
my_api_server.register_handler(OpCodes.ON_PLAYER_JOINING, my_custom_on_player_joining)

my_api_server.start()



import struct
from dataclasses import dataclass


@dataclass
class PlayerProgress:
    kill_count: int
    leader_kills: int
    assault_kills: int
    medic_kills: int
    engineer_kills: int
    support_kills: int
    recon_kills: int
    death_count: int
    win_count: int
    lose_count: int
    friendly_shots: int
    friendly_kills: int
    revived: int
    revived_team_mates: int
    assists: int
    prestige: int
    rank: int
    exp: int
    shots_fired: int
    shots_hit: int
    headshots: int
    objectives_completed: int
    healed_hp: int
    road_kills: int
    suicides: int
    vehicles_destroyed: int
    vehicle_hp_repaired: int
    longest_kill: int
    play_time_seconds: int
    leader_play_time: int
    assault_play_time: int
    medic_play_time: int
    engineer_play_time: int
    support_play_time: int
    recon_play_time: int
    leader_score: int
    assault_score: int
    medic_score: int
    engineer_score: int
    support_score: int
    recon_score: int
    total_score: int

    def __len__(self) -> int:
        """ Returns the number of attributes """
        return len(self.__dict__)

    def to_bytes(self) -> bytes:
        out = struct.pack("I", len(self))
        out += struct.pack("I", self.kill_count)
        out += struct.pack("I", self.leader_kills)
        out += struct.pack("I", self.assault_kills)
        out += struct.pack("I", self.medic_kills)
        out += struct.pack("I", self.engineer_kills)
        out += struct.pack("I", self.support_kills)
        out += struct.pack("I", self.recon_kills)
        out += struct.pack("I", self.death_count)
        out += struct.pack("I", self.win_count)
        out += struct.pack("I", self.lose_count)
        out += struct.pack("I", self.friendly_shots)
        out += struct.pack("I", self.friendly_kills)
        out += struct.pack("I", self.revived)
        out += struct.pack("I", self.revived_team_mates)
        out += struct.pack("I", self.assists)
        out += struct.pack("I", self.prestige)
        out += struct.pack("I", self.rank)
        out += struct.pack("I", self.exp)
        out += struct.pack("I", self.shots_fired)
        out += struct.pack("I", self.shots_hit)
        out += struct.pack("I", self.headshots)
        out += struct.pack("I", self.objectives_completed)
        out += struct.pack("I", self.healed_hp)
        out += struct.pack("I", self.road_kills)
        out += struct.pack("I", self.suicides)
        out += struct.pack("I", self.vehicles_destroyed)
        out += struct.pack("I", self.vehicle_hp_repaired)
        out += struct.pack("I", self.longest_kill)
        out += struct.pack("I", self.play_time_seconds)
        out += struct.pack("I", self.leader_play_time)
        out += struct.pack("I", self.assault_play_time)
        out += struct.pack("I", self.medic_play_time)
        out += struct.pack("I", self.engineer_play_time)
        out += struct.pack("I", self.support_play_time)
        out += struct.pack("I", self.recon_play_time)
        out += struct.pack("I", self.leader_score)
        out += struct.pack("I", self.assault_score)
        out += struct.pack("I", self.medic_score)
        out += struct.pack("I", self.engineer_score)
        out += struct.pack("I", self.support_score)
        out += struct.pack("I", self.recon_score)
        out += struct.pack("I", self.total_score)
        return out


def build_player_progress_from_bytes(buf: bytes) -> tuple[PlayerProgress, bytes]:
    """ Builds a PlayerProgress object from bytes and returns this alongside of the remaining, unread bytes. """
    parameter_count = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    kill_count = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    leader_kills = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    assault_kills = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    medic_kills = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    engineer_kills = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    support_kills = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    recon_kills = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    death_count = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    win_count = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    lose_count = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    friendly_shots = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    friendly_kills = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    revived = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    revived_team_mates = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    assists = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    prestige = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    rank = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    exp = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    shots_fired = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    shots_hit = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    headshots = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    objectives_completed = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    healed_hp = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    road_kills = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    suicides = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    vehicles_destroyed = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    vehicle_hp_repaired = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    longest_kill = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    play_time_seconds = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    leader_play_time = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    assault_play_time = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    medic_play_time = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    engineer_play_time = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    support_play_time = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    recon_play_time = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    leader_score = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    assault_score = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    medic_score = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    engineer_score = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    support_score = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    recon_score = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]
    total_score = struct.unpack("I", buf[:4])[0]
    buf = buf[4:]

    return PlayerProgress(kill_count=kill_count,
                          leader_kills=leader_kills,
                          assault_kills=assault_kills,
                          medic_kills=medic_kills,
                          engineer_kills=engineer_kills,
                          support_kills=support_kills,
                          recon_kills=recon_kills,
                          death_count=death_count,
                          win_count=win_count,
                          lose_count=lose_count,
                          friendly_shots=friendly_shots,
                          friendly_kills=friendly_kills,
                          revived=revived,
                          revived_team_mates=revived_team_mates,
                          assists=assists,
                          prestige=prestige,
                          rank=rank,
                          exp=exp,
                          shots_fired=shots_fired,
                          shots_hit=shots_hit,
                          headshots=headshots,
                          objectives_completed=objectives_completed,
                          healed_hp=healed_hp,
                          road_kills=road_kills,
                          suicides=suicides,
                          vehicles_destroyed=vehicles_destroyed,
                          vehicle_hp_repaired=vehicle_hp_repaired,
                          longest_kill=longest_kill,
                          play_time_seconds=play_time_seconds,
                          leader_play_time=leader_play_time,
                          assault_play_time=assault_play_time,
                          medic_play_time=medic_play_time,
                          engineer_play_time=engineer_play_time,
                          support_play_time=support_play_time,
                          recon_play_time=recon_play_time,
                          leader_score=leader_score,
                          assault_score=assault_score,
                          medic_score=medic_score,
                          engineer_score=engineer_score,
                          support_score=support_score,
                          recon_score=recon_score,
                          total_score=total_score), buf



import struct
from dataclasses import dataclass

from battlebit_community_server_api.helper.StructHelper import read_uint8
from battlebit_community_server_api.model.Gadget import read_from_bytes as build_gadget, Gadget
from battlebit_community_server_api.model.WeaponItem import read_from_bytes as build_weapon_item, WeaponItem


@dataclass
class PlayerLoadout:
    primary_weapon: WeaponItem
    secondary_weapon: WeaponItem

    first_aid: Gadget
    light_gadget: Gadget
    heavy_gadget: Gadget
    throwable: Gadget

    primary_extra_magazine: int
    secondary_extra_magazine: int
    first_aid_extra: int
    light_gadget_extra: int
    heavy_gadget_extra: int
    throwable_extra: int

    def as_bytes(self) -> bytes:
        out = self.primary_weapon.as_bytes()
        out += self.secondary_weapon.as_bytes()
        out += self.first_aid.as_bytes()
        out += self.light_gadget.as_bytes()
        out += self.heavy_gadget.as_bytes()
        out += self.throwable.as_bytes()
        out += struct.pack("B", self.primary_extra_magazine)
        out += struct.pack("B", self.secondary_extra_magazine)
        out += struct.pack("B", self.first_aid_extra)
        out += struct.pack("B", self.light_gadget_extra)
        out += struct.pack("B", self.heavy_gadget_extra)
        out += struct.pack("B", self.throwable_extra)
        return out


def read_from_bytes(data: bytes) -> tuple[PlayerLoadout, bytes]:
    """ Parses bytes and returns PlayerLoadout and remaining bytes """
    primary_weapon, data = build_weapon_item(data)
    secondary_weapon, data = build_weapon_item(data)
    first_aid, data = build_gadget(data)
    light_gadget, data = build_gadget(data)
    heavy_gadget, data = build_gadget(data)
    throwable, data = build_gadget(data)
    primary_mags, data = read_uint8(data)
    secondary_mags, data = read_uint8(data)
    first_aid_extra, data = read_uint8(data)
    light_gadget_extra, data = read_uint8(data)
    heavy_gadget_extra, data = read_uint8(data)
    throwable_extra, data = read_uint8(data)
    player_loadout = PlayerLoadout(primary_weapon=primary_weapon,
                                   secondary_weapon=secondary_weapon,
                                   first_aid=first_aid,
                                   light_gadget=light_gadget,
                                   heavy_gadget=heavy_gadget,
                                   throwable=throwable,
                                   primary_extra_magazine=primary_mags,
                                   secondary_extra_magazine=secondary_mags,
                                   first_aid_extra=first_aid_extra,
                                   light_gadget_extra=light_gadget_extra,
                                   heavy_gadget_extra=heavy_gadget_extra,
                                   throwable_extra=throwable_extra)
    return player_loadout, data



import logging
from typing import Optional

from battlebit_community_server_api.helper.StructHelper import *
from battlebit_community_server_api.model import PlayerProgress
from battlebit_community_server_api.model.ChatChannel import ChatChannel
from battlebit_community_server_api.model.GameRole import GameRole
from battlebit_community_server_api.model.MapSize import MapSize
from battlebit_community_server_api.model.OpCodes import OpCodes
from battlebit_community_server_api.model.GameState import GameState
from battlebit_community_server_api.model.PlayerLoadout import read_from_bytes as build_player_loadout, \
    PlayerLoadout
from battlebit_community_server_api.model.PlayerSpawnArguments import PlayerSpawnArguments
from battlebit_community_server_api.model.PlayerSpawningPosition import PlayerSpawningPosition
from battlebit_community_server_api.model.PlayerStand import PlayerStand
from battlebit_community_server_api.model.PlayerStats import PlayerStats
from battlebit_community_server_api.model.Role import Role
from battlebit_community_server_api.model.RoomSettings import RoomSettings
from battlebit_community_server_api.model.ServerInfo import ServerInfo
from battlebit_community_server_api.model.Squads import Squads
from battlebit_community_server_api.model.SteamId import SteamId
from battlebit_community_server_api.model.Team import Team
from battlebit_community_server_api.model.PlayerWearings import read_from_bytes as build_player_wearings, PlayerWearings
from battlebit_community_server_api.model.Vector3 import read_from_bytes as build_vector3


class TcpParsingService:
    """ The TcpParsingService parses TCP packages in their raw, bytes format and returns data objects. """
    ROUND_SETTINGS_SIZE: int = 33
    MAX_SQUADS_PER_TEAM: int = 64

    @staticmethod
    def split_data_into_individual_messages(data: bytes) -> tuple[tuple[OpCodes, bytes]]:
        """ Takes a bytestring of any length, splits it in individual messages and returns them with OpCodes """
        messages = []
        while len(data) > 0:
            packet_size, data = read_uint32(data)
            message = data[:packet_size]
            data = data[packet_size:]
            op_code_raw, message = read_uint8(message)
            messages.append((OpCodes(op_code_raw), message))
        return tuple(messages)

    @staticmethod
    def parse_on_player_connected(message: bytes) -> None:  # FixMe: UNTESTED
        steam_id = struct.unpack("Q", message[:8])[0]
        message = message[8:]
        user_name_size = struct.unpack("H", message[:2])[0]
        message = message[2:]
        if user_name_size != 0:  # ToDo: Parse this as soon as data comes available, set buffer position accordingly
            print("FOUND NEW DATASET, USERNAME SIZE != 0\n", message)
        ip_address = struct.unpack("I", message[:4])[0]
        message = message[4:]
        team = int(struct.unpack("B", message[:1])[0])
        message = message[1:]
        squad = int(struct.unpack("B", message[:1])[0])
        message = message[1:]
        role = int(struct.unpack("B", message[:1])[0])
        message = message[1:]

    @staticmethod
    def parse_on_player_typed_message(message: bytes) -> tuple[int, SteamId, ChatChannel, str]:
        """ Returns tuple with messageID, SteamID, ChatChannel and the message string """
        message_id, message = read_uint16(message)
        steam_id, message = read_uint64(message)
        channel, message = read_uint8(message)
        message_length, message = read_uint16(message)
        return message_id, SteamId(steam_id), ChatChannel(channel), message[:message_length].decode("utf-8")

    @staticmethod
    def parse_on_player_joining(message: bytes) -> tuple[SteamId, PlayerStats]:
        """ Returns a tuple with SteamID and PlayerStats """
        steam_id, message = read_uint64(message)
        steam_id = SteamId(steam_id)
        is_banned, message = read_uint8(message)
        is_banned = bool(is_banned)
        roles, message = read_uint64(message)

        progress, message = PlayerProgress.build_player_progress_from_bytes(message)

        tool_progress_size, message = read_uint16(message)
        tool_progress_raw = message[:tool_progress_size]
        message = message[tool_progress_size:]

        achievements_size, message = read_uint16(message)
        achievements_raw = message[:achievements_size]
        message = message[achievements_size:]

        selections_size, message = read_uint16(message)
        selections_raw = message[:selections_size]
        message = message[selections_size:]

        return steam_id, PlayerStats(is_banned=is_banned,
                                     roles=roles,
                                     progress=progress,
                                     tool_progress=tool_progress_raw,
                                     achievements=achievements_raw,
                                     selections=selections_raw)

    @classmethod
    def parse_hail_message(cls, message: bytes) -> ServerInfo:
        """ Parses HAIL """
        # ToDo: Parse Map/GameMode Rotation and Room Settings and add them to ServerInfo

        # Token
        token_size, message = read_uint16(message)
        token, message = read_string(message, token_size)

        # Version
        version_size, message = read_uint16(message)
        version, message = read_string(message, version_size)

        # Game Port
        game_port, message = read_uint16(message)

        # is port protected
        is_port_protected = struct.unpack("?", message[:1])[0]
        message = message[1:]

        # length of server name
        server_name_length, message = read_uint16(message)

        # Server Name
        server_name = message[:server_name_length].decode("utf-8")
        message = message[server_name_length:]

        # length of game mode name
        game_mode_name_length, message = read_uint16(message)

        # Game Mode Name
        game_mode_name = message[:game_mode_name_length].decode("utf-8")
        message = message[game_mode_name_length:]

        # length of map name
        map_name_length, message = read_uint16(message)

        # Map Name
        map_name = message[:map_name_length].decode("utf-8")
        message = message[map_name_length:]

        # Map Size
        map_size, message = read_uint8(message)

        # DayNight Cycle
        day_night_cycle, message = read_uint8(message)

        # Current Players
        current_players, message = read_uint8(message)

        # Queued Players
        queued_players, message = read_uint8(message)

        # Max Players
        max_players, message = read_uint8(message)

        # Loading screen text length
        loading_screen_text_length, message = read_uint16(message)

        # Loading screen text
        loading_screen_text = message[:loading_screen_text_length].decode("utf-8")
        message = message[loading_screen_text_length:]

        # Server Rules text size
        server_rules_text_length, message = read_uint16(message)

        # Server rules text
        server_rules_text = message[:server_rules_text_length].decode("utf-8")
        message = message[server_rules_text_length:]

        # Round Index
        round_index, message = read_uint32(message)

        # Session ID
        session_id, message = read_uint64(message)

        # Room Settings size
        room_settings_length, message = read_uint32(message)

        # Room Settings
        cls._parse_room_settings(message[:room_settings_length])
        message = message[room_settings_length:]

        # Map & Game mode Rotation size
        map_and_game_mode_rotation_size, message = read_uint32(message)

        # Map & Game mode Rotation
        cls.parse_map_game_mode_rotation(message[:map_and_game_mode_rotation_size])
        message = message[map_and_game_mode_rotation_size:]

        # Round Settings
        round_settings = cls._parse_round_settings(message[:cls.ROUND_SETTINGS_SIZE])
        message = message[cls.ROUND_SETTINGS_SIZE:]

        # Clients
        client_count, message = read_uint8(message)
        clients = []
        for i in range(client_count):
            steam_id, user_name, ip_hash, team, squad, role, is_alive, loadout, wearings, message = \
                cls.parse_hail_message_client_info(message)
            clients.append((steam_id, user_name, ip_hash, team, squad, role, is_alive, loadout, wearings))

        logging.debug("Connected clients:")
        for client in clients:
            logging.debug(f"[{client[0]}] {client[1]}")

        # Squads
        squad_data_size, message = read_uint32(message)
        squad_points_team_us = []
        for _ in range(cls.MAX_SQUADS_PER_TEAM):
            points, message = read_uint32(message)
            squad_points_team_us.append(points)

        squad_points_team_ru = []
        for _ in range(cls.MAX_SQUADS_PER_TEAM):
            points, message = read_uint32(message)
            squad_points_team_ru.append(points)

        if message:
            logging.warning(f"Unparsed HAIL message remains: {message}. "
                            f"Probably an update that broke everything. Please fix.")

        return ServerInfo(name=server_name,
                          version=version,
                          port=game_port,
                          protected=is_port_protected,
                          game_mode=game_mode_name,
                          map_name=map_name,
                          map_size=MapSize(map_size),
                          is_day_mode=day_night_cycle == 0,
                          current_players=current_players,
                          queued_players=queued_players,
                          max_players=max_players,
                          loading_screen_text=loading_screen_text,
                          rules_text=server_rules_text,
                          game_state=GameState(round_settings[0]),
                          seconds_left_in_round=round_settings[-1])

    @staticmethod
    def _parse_room_settings(room_settings_bytes: bytes) -> RoomSettings:
        damage_multiplier, room_settings_bytes = read_float32(room_settings_bytes)
        friendly_fire_enabled, room_settings_bytes = read_bool(room_settings_bytes)
        hide_map_votes, room_settings_bytes = read_bool(room_settings_bytes)
        only_winner_team_can_vote, room_settings_bytes = read_bool(room_settings_bytes)
        player_collision, room_settings_bytes = read_bool(room_settings_bytes)

        medic_limit_per_squad, room_settings_bytes = read_uint8(room_settings_bytes)
        engineer_limit_per_squad, room_settings_bytes = read_uint8(room_settings_bytes)
        support_limit_per_squad, room_settings_bytes = read_uint8(room_settings_bytes)
        recon_limit_per_squad, room_settings_bytes = read_uint8(room_settings_bytes)

        return RoomSettings(damage_multiplier=damage_multiplier,
                            friendly_fire_enabled=friendly_fire_enabled,
                            hide_map_votes=hide_map_votes,
                            only_winner_team_can_vote=only_winner_team_can_vote,
                            medic_limit_per_squad=medic_limit_per_squad,
                            engineer_limit_per_squad=engineer_limit_per_squad,
                            support_limit_per_squad=support_limit_per_squad,
                            recon_limit_per_squad=recon_limit_per_squad)

    @classmethod
    def parse_on_player_requesting_to_spawn(cls, message: bytes) -> tuple[SteamId, PlayerSpawnArguments, int]:
        """ Returns STEAM-ID, PlayerSpawnArguments and a Vehicle-ID """
        steam_id, message = read_uint64(message)
        steam_id = SteamId(steam_id)
        player_spawning_position, message = read_uint8(message)
        player_spawning_position = PlayerSpawningPosition(player_spawning_position)

        # LOADOUT
        player_loadout, message = build_player_loadout(message)

        # WEARINGS
        player_wearings, message = build_player_wearings(message)
        spawn_position, message = build_vector3(message)
        look_direction, message = build_vector3(message)
        stand, message = read_uint8(message)
        stand = PlayerStand(stand)
        spawn_protection, message = read_float32(message)

        # VEHICLE ID
        vehicle_id, message = read_uint16(message)

        return steam_id, PlayerSpawnArguments(
            position=player_spawning_position,
            loadout=player_loadout,
            wearings=player_wearings,
            vector_position=spawn_position,
            look_direction=look_direction,
            stand=stand,
            spawn_protection=spawn_protection
        ), vehicle_id

    @staticmethod
    def parse_on_player_asking_to_change_role(data: bytes) -> tuple[SteamId, GameRole]:
        steam_id, data = read_uint64(data)
        steam_id = SteamId(steam_id)
        requested_role, data = read_uint8(data)
        requested_role = GameRole(requested_role)
        return steam_id, requested_role

    @staticmethod
    def parse_on_player_asking_to_change_team(data: bytes) -> tuple[SteamId, Team]:
        steam_id, data = read_uint64(data)
        steam_id = SteamId(steam_id)
        requested_team, data = read_uint8(data)
        requested_team = Team(requested_team)
        return steam_id, requested_team

    @staticmethod
    def parse_map_game_mode_rotation(map_game_mode_rotation_bytes: bytes) -> None:
        # FixMe: Not clear how those values are to be understood
        map_rotation_size = struct.unpack("I", map_game_mode_rotation_bytes[:4])[0]  # Map Rotation
        game_mode_rotation_size = struct.unpack("I", map_game_mode_rotation_bytes[:4])[0]  # Game Mode Rotation

    @staticmethod
    def _parse_round_settings(round_settings_bytes: bytes) -> tuple[int, float, float, float, int, int]:
        state = struct.unpack('b', round_settings_bytes[0:1])[0]
        team_a_tickets = struct.unpack('d', round_settings_bytes[1:9])[0]
        team_b_tickets = struct.unpack('d', round_settings_bytes[9:17])[0]
        max_tickets = struct.unpack('d', round_settings_bytes[17:25])[0]
        players_to_start = struct.unpack('i', round_settings_bytes[25:29])[0]
        seconds_left = struct.unpack('i', round_settings_bytes[29:33])[0]
        return state, team_a_tickets, team_b_tickets, max_tickets, players_to_start, seconds_left

    @staticmethod
    def parse_hail_message_client_info(client_info: bytes) -> tuple[
        SteamId, str, int, Team, Squads, Role, bool, Optional[PlayerLoadout], Optional[PlayerWearings], bytes]:
        """
        Parses the guaranteed 24 bytes segments of client info, returned by the HAIL message.
        Then, if the player is alive, parses the PlayerLoadout and PlayerWearing.

        :return: Tuple containing: Steam-ID, Username, IP Hash, Team, Squad, Role, IsAliveStatus, PlayerLoadout,
                                   PlayerWearing and the remainder of buffer.
        :note: If the player is not alive, PlayerLoadout and PlayerWearing are None.
        """
        steam_id, client_info = read_uint64(client_info)
        steam_id = SteamId(steam_id)
        user_name_size, client_info = read_uint16(client_info)
        user_name, client_info = read_string(client_info, user_name_size)
        ip_hash, client_info = read_uint32(client_info)
        team, client_info = read_uint8(client_info)
        team = Team(team)
        squad, client_info = read_uint8(client_info)
        squad = Squads(squad)
        role, client_info = read_uint8(client_info)
        role = Role(role)
        is_alive, client_info = read_uint8(client_info)
        is_alive = bool(is_alive)
        if is_alive:
            _, client_info = read_uint32(client_info)  # Loadout size
            player_loadout, client_info = build_player_loadout(client_info)
            player_wearings, client_info = build_player_wearings(client_info)
        else:
            player_loadout = None
            player_wearings = None

        return steam_id, user_name, ip_hash, team, squad, role, is_alive, player_loadout, player_wearings, client_info



import struct
from dataclasses import dataclass

from battlebit_community_server_api.helper.StructHelper import read_float32


@dataclass
class Vector3:
    x: float
    y: float
    z: float

    def as_bytes(self) -> bytes:
        out = struct.pack("f", self.x)
        out += struct.pack("f", self.y)
        out += struct.pack("f", self.z)
        return out


def read_from_bytes(data: bytes) -> tuple[Vector3, bytes]:
    """ reads from zero, x => y => z, then returns Vector3 and remaining bytes """
    x, data = read_float32(data)
    y, data = read_float32(data)
    z, data = read_float32(data)
    return Vector3(x=x, y=y, z=z), data



from enum import IntEnum


class ReasonOfDamage(IntEnum):
    SERVER = 0
    WEAPON = 1
    BLEEDING = 2
    FALL = 3
    HELICOPTER_BLADE = 4
    VEHICLE_EXPLOSION = 5
    EXPLOSION = 6
    VEHICLE_RUNOVER = 7
    BUILDING_COLLAPSING = 8
    SLEDGEHAMMER = 9
    TREE_FALL = 10
    COUNT_AS_KILL = 11
    SUICIDE = 12
    HELICOPTER_CRASH = 13
    BARBED_WIRE = 14



import struct
from dataclasses import dataclass

from battlebit_community_server_api.model.PlayerStats import PlayerStats
from battlebit_community_server_api.model.Squads import Squads
from battlebit_community_server_api.model.Team import Team


@dataclass
class PlayerJoiningArguments:
    player_stats: PlayerStats
    team: Team
    squad: Squads

    def to_bytes(self) -> bytes:
        out = self.player_stats.to_bytes()
        out += struct.pack("B", self.team.value)
        out += struct.pack("B", self.squad.value)
        return out



from dataclasses import dataclass


@dataclass
class RoomSettings:
    damage_multiplier: float
    friendly_fire_enabled: bool
    hide_map_votes: bool
    only_winner_team_can_vote: bool

    medic_limit_per_squad: int
    engineer_limit_per_squad: int
    support_limit_per_squad: int
    recon_limit_per_squad: int



from dataclasses import dataclass

from battlebit_community_server_api.helper.StructHelper import write_string
from battlebit_community_server_api.model.AttachmentType import AttachmentType


@dataclass
class Attachment:
    name: str
    attachment_type: AttachmentType

    def as_bytes(self) -> bytes:
        return write_string(self.name)


# Barrels
BASIC = Attachment("Basic", AttachmentType.BARREL)
COMPENSATOR = Attachment("Compensator", AttachmentType.BARREL)
HEAVY = Attachment("Heavy", AttachmentType.BARREL)
LONG_BARREL = Attachment("Long_Barrel", AttachmentType.BARREL)
MUZZLE_BREAK = Attachment("Muzzle_Break", AttachmentType.BARREL)
RANGER = Attachment("Ranger", AttachmentType.BARREL)
SUPPRESSOR_LONG = Attachment("Suppressor_Long", AttachmentType.BARREL)
SUPPRESSOR_SHORT = Attachment("Suppressor_Short", AttachmentType.BARREL)
TACTICAL = Attachment("Tactical", AttachmentType.BARREL)
FLASH_HIDER = Attachment("Flash_Hider", AttachmentType.BARREL)
OSPREY_9 = Attachment("Osprey_9", AttachmentType.BARREL)
DGN_308 = Attachment("DGN-308", AttachmentType.BARREL)
VAMB_762 = Attachment("VAMB-762", AttachmentType.BARREL)
SDN_6762 = Attachment("SDN-6_762", AttachmentType.BARREL)
NT_4556 = Attachment("NT-4_556", AttachmentType.BARREL)

# Canted Sights
IRON_SIGHT = Attachment("Ironsight", AttachmentType.CANTED_SIGHT)
CANTED_RED_DOT = Attachment("Canted_Red_Dot", AttachmentType.CANTED_SIGHT)
FYOU_CANTED = Attachment("FYou_Canted", AttachmentType.CANTED_SIGHT)
HOLO_DOT = Attachment("Holo_Dot", AttachmentType.CANTED_SIGHT)

# Scope
_6X_SCOPE = Attachment("6x_Scope", AttachmentType.MAIN_SIGHT)
_8X_SCOPE = Attachment("8x_Scope", AttachmentType.MAIN_SIGHT)
_15X_SCOPE = Attachment("15x_Scope", AttachmentType.MAIN_SIGHT)
_20X_SCOPE = Attachment("20x_Scope", AttachmentType.MAIN_SIGHT)
PTR_40_HUNTER = Attachment("PTR-40_Hunter", AttachmentType.MAIN_SIGHT)
_1P78 = Attachment("1P78", AttachmentType.MAIN_SIGHT)
ACOG = Attachment("Acog", AttachmentType.MAIN_SIGHT)
M125 = Attachment("M_125", AttachmentType.MAIN_SIGHT)
PRISMA = Attachment("Prisma", AttachmentType.MAIN_SIGHT)
SLIP = Attachment("Slip", AttachmentType.MAIN_SIGHT)
PISTOL_DELTA_SIGHT = Attachment("Pistol_Delta_Sight", AttachmentType.MAIN_SIGHT)
PISTOL_RED_DOT = Attachment("Pistol_Red_Dot", AttachmentType.MAIN_SIGHT)
AIM_COMP = Attachment("Aim_Comp", AttachmentType.MAIN_SIGHT)
HOLOGRAPHIC = Attachment("Holographic", AttachmentType.MAIN_SIGHT)
KOBRA = Attachment("Kobra", AttachmentType.MAIN_SIGHT)
OKP7 = Attachment("OKP7", AttachmentType.MAIN_SIGHT)
PKAS = Attachment("PK-AS", AttachmentType.MAIN_SIGHT)
RED_DOT = Attachment("Red_Dot", AttachmentType.MAIN_SIGHT)
REFLEX = Attachment("Reflex", AttachmentType.MAIN_SIGHT)
STRIKEFIRE = Attachment("Strikefire", AttachmentType.MAIN_SIGHT)
RAZOR = Attachment("Razor", AttachmentType.MAIN_SIGHT)
FLIR = Attachment("Flir", AttachmentType.MAIN_SIGHT)
ECHO = Attachment("Echo", AttachmentType.MAIN_SIGHT)
TRI4X32 = Attachment("TRI4X32", AttachmentType.MAIN_SIGHT)
FYOU_SIGHT = Attachment("FYou_Sight", AttachmentType.MAIN_SIGHT)
HOLO_PK120 = Attachment("Holo_PK-120", AttachmentType.MAIN_SIGHT)
PISTOL_8X_SCOPE = Attachment("Pistol_8x_Scope", AttachmentType.MAIN_SIGHT)
BURRIS_AR332 = Attachment("BurrisAR332", AttachmentType.MAIN_SIGHT)
HS401G5 = Attachment("HS401G5", AttachmentType.MAIN_SIGHT)

# Top Scope
DELTA_SIGHT_TOP = Attachment("Delta_Sight_Top", AttachmentType.TOP_SIGHT)
RED_DOT_TOP = Attachment("Red_Dot_Top", AttachmentType.TOP_SIGHT)
C_Red_DOT_TOP = Attachment("C_Red_Dot_Top", AttachmentType.TOP_SIGHT)
FYOU_TOP = Attachment("FYou_Top", AttachmentType.TOP_SIGHT)

# Under Rails
ANGLED_GRIP = Attachment("Angled_Grip", AttachmentType.UNDER_RAIL)
BIPOD = Attachment("Bipod", AttachmentType.UNDER_RAIL)
VERTICAL_GRIP = Attachment("Vertical_Grip", AttachmentType.UNDER_RAIL)
STUBBY_GRIP = Attachment("Stubby_Grip", AttachmentType.UNDER_RAIL)
STABIL_GRIP = Attachment("Stabil_Grip", AttachmentType.UNDER_RAIL)
VERTICAL_SKELETON_GRIP = Attachment("Vertical_Skeleton_Grip", AttachmentType.UNDER_RAIL)
FABDTFG = Attachment("FAB-DTFG", AttachmentType.UNDER_RAIL)
MAGPUL_ANGLED = Attachment("Magpul_Angled", AttachmentType.UNDER_RAIL)
BCMGUN_FIGHTER = Attachment("BCM-Gun_Fighter", AttachmentType.UNDER_RAIL)
SHIFT_SHORT_ANGLED_GRIP = Attachment("Shift_Short_Angled_Grip", AttachmentType.UNDER_RAIL)
SE5_GRIP = Attachment("SE-5_Grip", AttachmentType.UNDER_RAIL)
RK6_FORE_GRIP = Attachment("RK-6_Foregrip", AttachmentType.UNDER_RAIL)
HERA_CQR_FRONT = Attachment("HeraCQR_Front", AttachmentType.UNDER_RAIL)
B25URK = Attachment("B-25URK", AttachmentType.UNDER_RAIL)
VTAC_UVG_TACTICAL_GRIP = Attachment("VTAC_UVG_TacticalGrip", AttachmentType.UNDER_RAIL)

# Side Rails
FLASHLIGHT = Attachment("Flashlight", AttachmentType.SIDE_RAIL)
RANGEFINDER = Attachment("Rangefinder", AttachmentType.SIDE_RAIL)
RED_LASER = Attachment("Redlaser", AttachmentType.SIDE_RAIL)
TACTICAL_FLASHLIGHT = Attachment("Tactical_Flashlight", AttachmentType.SIDE_RAIL)
GREEN_LASER = Attachment("Greenlaser", AttachmentType.SIDE_RAIL)
SEARCH_LIGHT = Attachment("Searchlight", AttachmentType.SIDE_RAIL)

# Bolts
BOLT_ACTION_A = Attachment("BOLT_Action_A", AttachmentType.BOLT)
BOLT_ACTION_B = Attachment("BOLT_Action_B", AttachmentType.BOLT)
BOLT_ACTION_C = Attachment("BOLT_Action_C", AttachmentType.BOLT)
BOLT_ACTION_D = Attachment("BOLT_Action_D", AttachmentType.BOLT)
BOLT_ACTION_E = Attachment("BOLT_Action_E", AttachmentType.BOLT)

ALL_ATTACHMENTS = {a.name: a for a in [BASIC, COMPENSATOR, HEAVY, LONG_BARREL, MUZZLE_BREAK, RANGER, SUPPRESSOR_LONG,
                                       SUPPRESSOR_SHORT, TACTICAL, FLASH_HIDER, OSPREY_9, DGN_308, VAMB_762, SDN_6762,
                                       NT_4556, IRON_SIGHT, CANTED_RED_DOT, FYOU_CANTED, HOLO_DOT, _6X_SCOPE,
                                       _8X_SCOPE, _15X_SCOPE, _20X_SCOPE, PTR_40_HUNTER, _1P78, ACOG, M125, PRISMA,
                                       SLIP, PISTOL_DELTA_SIGHT, PISTOL_RED_DOT, AIM_COMP, HOLOGRAPHIC, KOBRA, OKP7,
                                       PKAS, RED_DOT, REFLEX, STRIKEFIRE, RAZOR, FLIR, ECHO, TRI4X32, FYOU_SIGHT,
                                       HOLO_PK120, PISTOL_8X_SCOPE, BURRIS_AR332, HS401G5, DELTA_SIGHT_TOP,
                                       RED_DOT_TOP, C_Red_DOT_TOP, FYOU_TOP, ANGLED_GRIP, BIPOD, VERTICAL_GRIP,
                                       STUBBY_GRIP, STABIL_GRIP, VERTICAL_SKELETON_GRIP, FABDTFG, MAGPUL_ANGLED,
                                       BCMGUN_FIGHTER, SHIFT_SHORT_ANGLED_GRIP, SE5_GRIP, RK6_FORE_GRIP,
                                       HERA_CQR_FRONT, B25URK, VTAC_UVG_TACTICAL_GRIP, FLASHLIGHT, RANGEFINDER,
                                       RED_LASER, TACTICAL_FLASHLIGHT, GREEN_LASER, SEARCH_LIGHT, BOLT_ACTION_A,
                                       BOLT_ACTION_B, BOLT_ACTION_C, BOLT_ACTION_D, BOLT_ACTION_E]}



from enum import IntEnum


class OpCodes(IntEnum):
    NONE = 0
    HAIL = 1
    ACCEPTED = 2
    DENIED = 3

    EXECUTE_COMMAND = 10
    SEND_PLAYER_STATS = 11
    SPAWN_PLAYER = 12
    SET_NEW_ROOM_SETTINGS = 13
    RESPOND_PLAYER_MESSAGE = 14
    SET_NEW_ROUND_STATE = 15
    SET_PLAYER_WEAPON = 16
    SET_PLAYER_GADGET = 17

    PLAYER_CONNECTED = 50
    PLAYER_DISCONNECTED = 51
    ON_PLAYER_TYPED_MESSAGE = 52
    ON_PLAYER_KILLED_ANOTHER_PLAYER = 53
    ON_PLAYER_JOINING = 54
    SAVE_PLAYER_STATS = 55
    ON_PLAYER_ASKING_TO_CHANGE_ROLE = 56
    ON_PLAYER_CHANGED_ROLE = 57
    ON_PLAYER_JOINED_A_SQUAD = 58
    ON_PLAYER_LEFT_SQUAD = 59
    ON_PLAYER_CHANGED_TEAM = 60
    ON_PLAYER_REQUESTING_TO_SPAWN = 61
    ON_PLAYER_REPORT = 62
    ON_PLAYER_SPAWN = 63
    ON_PLAYER_DIE = 64
    NOTIFY_NEW_MAP_ROTATION = 65
    NOTIFY_NEW_GAME_MODE_ROTATION = 66
    NOTIFY_NEW_ROUND_STATE = 67
    ON_PLAYER_ASKING_TO_CHANGE_TEAM = 68
    GAME_TICK = 69
    ON_PLAYER_GIVEN_UP = 70
    ON_PLAYER_REVIVED_ANOTHER = 71
    ON_SQUAD_POINTS_CHANGED = 72
    NOTIFY_NEW_ROUND_ID = 73
    LOG = 74
    ON_SQUAD_LEADER_CHANGED = 75
    UPDATE_NEW_GAME_DATA = 76
    UPDATE_CONNECTED_PLAYERS = 77



from dataclasses import dataclass
from typing import Optional

from battlebit_community_server_api.helper.StructHelper import *


@dataclass
class Gadget:
    name: str

    def as_bytes(self) -> bytes:
        return write_string(self.name)


def read_from_bytes(data: bytes) -> tuple[Optional[Gadget], bytes]:
    """ Reads from 0 until data is completed, returns remaining bytes """
    gadget_size, data = read_uint16(data)
    gadget_name, data = read_string(data, gadget_size)
    try:
        gadget = ALL_GADGETS[gadget_name]
    except KeyError:
        gadget = None
    return gadget, data


BANDAGE = Gadget("Bandage")
BINOCULARS = Gadget("Binoculars")
RANGE_FINDER = Gadget("Range Finder")
REPAIR_TOOL = Gadget("Repair Tool")
C4 = Gadget("C4")
CLAYMORE = Gadget("Claymore")
M320_SMOKE_GRENADE_LAUNCHER = Gadget("M320 Smoke Grenade Launcher")
SMALL_AMMO_KIT = Gadget("Small Ammo Kit")
ANTI_PERSONNEL_MINE = Gadget("Anti Personnel Mine")
ANTI_VEHICLE_MINE = Gadget("Anti Vehicle Mine")
MEDIC_KIT = Gadget("Medic Kit")
RPG7_HEAT_EXPLOSIVE = Gadget("Rpg7 Heat Explosive")
RIOT_SHIELD = Gadget("Riot Shield")
FRAG_GRENADE = Gadget("Frag Grenade")
IMPACT_GRENADE = Gadget("Impact Grenade")
ANTI_VEHICLE_GRENADE = Gadget("Anti Vehicle Grenade")
SMOKE_GRENADE_BLUE = Gadget("Smoke Grenade Blue")
SMOKE_GRENADE_GREEN = Gadget("Smoke Grenade Green")
SMOKE_GRENADE_RED = Gadget("Smoke Grenade Red")
SMOKE_GRENADE_WHITE = Gadget("Smoke Grenade White")
FLARE = Gadget("Flare")
SLEDGE_HAMMER = Gadget("Sledge Hammer")
ADVANCED_BINOCULARS = Gadget("Advanced Binoculars")
MDX201 = Gadget("Mdx 201")
BINO_SOFLAM = Gadget("Bino Soflam")
HEAVY_AMMO_KIT = Gadget("Heavy Ammo Kit")
RPG7_PGO7_TANDEM = Gadget("Rpg7 Pgo7 Tandem")
RPG7_PGO7_HEAT_EXPLOSIVE = Gadget("Rpg7 Pgo7 Heat Explosive")
RPG7_PGO7_FRAGMENTATION = Gadget("Rpg7 Pgo7 Fragmentation")
RPG7_FRAGMENTATION = Gadget("Rpg7 Fragmentation")
GRAPPLING_HOOK = Gadget("Grappling Hook")
AIR_DRONE = Gadget("Air Drone")
FLASHBANG = Gadget("Flashbang")
PICKAXE = Gadget("Pickaxe")
SUICIDE_C4 = Gadget("SuicideC4")
SLEDGE_HAMMER_SKIN_A = Gadget("Sledge Hammer SkinA")
SLEDGE_HAMMER_SKIN_B = Gadget("Sledge Hammer SkinB")
SLEDGE_HAMMER_SKIN_C = Gadget("Sledge Hammer SkinC")
PICKAXE_IRON_PICKAXE = Gadget("Pickaxe IronPickaxe")

ALL_GADGETS = {g.name: g for g in [
    BANDAGE,
    BINOCULARS,
    RANGE_FINDER,
    REPAIR_TOOL,
    C4,
    CLAYMORE,
    M320_SMOKE_GRENADE_LAUNCHER,
    SMALL_AMMO_KIT,
    ANTI_PERSONNEL_MINE,
    ANTI_VEHICLE_MINE,
    MEDIC_KIT,
    RPG7_HEAT_EXPLOSIVE,
    RIOT_SHIELD,
    FRAG_GRENADE,
    IMPACT_GRENADE,
    ANTI_VEHICLE_GRENADE,
    SMOKE_GRENADE_BLUE,
    SMOKE_GRENADE_GREEN,
    SMOKE_GRENADE_RED,
    SMOKE_GRENADE_WHITE,
    FLARE,
    SLEDGE_HAMMER,
    ADVANCED_BINOCULARS,
    MDX201,
    BINO_SOFLAM,
    HEAVY_AMMO_KIT,
    RPG7_PGO7_TANDEM,
    RPG7_PGO7_HEAT_EXPLOSIVE,
    RPG7_PGO7_FRAGMENTATION,
    RPG7_FRAGMENTATION,
    GRAPPLING_HOOK,
    AIR_DRONE,
    FLASHBANG,
    PICKAXE,
    SUICIDE_C4,
    SLEDGE_HAMMER_SKIN_A,
    SLEDGE_HAMMER_SKIN_B,
    SLEDGE_HAMMER_SKIN_C,
    PICKAXE_IRON_PICKAXE
]}



from dataclasses import dataclass

from battlebit_community_server_api.model.GameState import GameState
from battlebit_community_server_api.model.MapSize import MapSize


@dataclass
class ServerInfo:
    name: str
    version: str
    port: int
    protected: bool
    game_mode: str
    map_name: str
    map_size: MapSize
    is_day_mode: bool
    current_players: int
    queued_players: int
    max_players: int
    loading_screen_text: str
    rules_text: str
    game_state: GameState
    seconds_left_in_round: int

    def generate_log(self) -> str:
        return f"Server: {self.name}\n" \
               f"Version: {self.version}\n" \
               f"Port: {self.port} ({'' if self.protected else 'un'}protected)\n" \
               f"Game Mode: {self.game_mode}\n" \
               f"Map: {self.map_name} (Size: {MapSize(self.map_size).name.replace('_', ' ').strip()})\n" \
               f"Day Mode: {'Yes' if self.is_day_mode == 0 else 'No'}\n" \
               f"Players (current/in queue/max): {self.current_players}/{self.queued_players}/{self.max_players}\n" \
               f"Loading Screen Text: {self.loading_screen_text}\n" \
               f"Rules: {self.rules_text}\n" \
               f"Game State: {GameState(self.game_state).name.replace('_', ' ')}\n" \
               f"Seconds left: {self.seconds_left_in_round}"
